@@ -58,7 +58,7 @@ class ExceptionInfo:
         return f"{self.type}, ({hex(self.tb_start)}, {hex(self.tb_size)}, {self.tb_icount})"
 
 class UnicornPageManager(PageManager):
-    def __init__(self, uc: Uc):
+    def __init__(self, uc: Uc) -> None:
         self._uc = uc
 
     def commit(self, addr: int, size: int, protect: MemoryProtect) -> None:
@@ -71,10 +71,185 @@ class UnicornPageManager(PageManager):
     def protect(self, addr: int, size: int, protect: MemoryProtect) -> None:
         self._uc.mem_protect(addr, size, map_unicorn_perms(protect))
 
+    def read(self, addr: int, size: int) -> bytearray:
+        return self._uc.mem_read(addr, size)
+
+    def write(self, addr: int, data: bytes) -> None:
+        if not isinstance(data, bytes):
+            data = bytes(data)
+        self._uc.mem_write(addr, data)
+
+class LazyPage:
+    def __init__(self, addr: int, protect: MemoryProtect, committed: bool):
+        self.addr = addr
+        self.protect = protect
+        self.committed = committed
+        self.data: Optional[bytearray] = None
+
+    @property
+    def size(self):
+        return PAGE_SIZE
+
+class LazyPageManager(PageManager):
+    def __init__(self, child: PageManager):
+        self.child = child
+        self.total_commit = 0
+        self.pages: Dict[int, LazyPage] = {}
+        self.lazy = True
+
+    def iter_pages(self, addr: int, size: int):
+        for i in range(0, size // PAGE_SIZE):
+            page_addr = addr + i * PAGE_SIZE
+            yield page_addr
+
+    def iter_chunks(self, addr: int, size: int):
+        # TODO: rewrite this to not be so disgusting
+        page = addr & ~0xFFF
+        index = addr & 0xFFF
+        while True:
+            if page >= addr + size:
+                break
+            length = min(PAGE_SIZE, (addr + size) - (page + index))
+            yield page, index, length
+            page += PAGE_SIZE
+            index = 0
+
+    def handle_lazy_page(self, addr: int, size: int) -> bool:
+        try:
+            result = False
+            for page_addr, index, length in self.iter_chunks(addr, size):
+                page = self.pages.get(page_addr, None)
+                if page is None:
+                    continue
+                if not page.committed:
+                    self.child.commit(page.addr, page.size, page.protect)
+                    page.committed = True
+                    if page.data is not None:
+                        self.child.write(page.addr, page.data)
+                        page.data = None
+                    result = True
+            return result
+        except UcError as err:
+            print(f"FATAL ERROR {err}: handle_lazy_page({hex(addr)}[{hex(size)}])")
+            return False
+
+    def commit(self, addr: int, size: int, protect: MemoryProtect) -> None:
+        assert addr & 0xFFF == 0
+        assert size & 0xFFF == 0
+        #print(f"commit({hex(addr)}, {hex(size)}, {protect})")
+        if not self.lazy:
+            self.child.commit(addr, size, protect)
+        for page_addr in self.iter_pages(addr, size):
+            assert page_addr not in self.pages
+            self.pages[page_addr] = LazyPage(page_addr, protect, not self.lazy)
+        self.total_commit += size
+
+    def decommit(self, addr: int, size: int) -> None:
+        assert addr & 0xFFF == 0
+        assert size & 0xFFF == 0
+        #print(f"decommit({hex(addr)}, {hex(size)})")
+        pages = []
+        for page_addr in self.iter_pages(addr, size):
+            assert page_addr in self.pages
+            pages.append(self.pages[page_addr])
+
+        if all(page.committed for page in pages):
+            self.child.decommit(addr, size)
+        else:
+            for page in pages:
+                if page.committed:
+                    self.child.decommit(page.addr, page.size)
+
+        for page_addr in self.iter_pages(addr, size):
+            del self.pages[page_addr]
+
+    def protect(self, addr: int, size: int, protect: MemoryProtect) -> None:
+        assert addr & 0xFFF == 0
+        assert size & 0xFFF == 0
+        #print(f"protect({hex(addr)}, {hex(size)}, {protect})")
+        pages = []
+        for page_addr in self.iter_pages(addr, size):
+            assert page_addr in self.pages
+            pages.append(self.pages[page_addr])
+
+        if all(page.committed for page in pages):
+            self.child.protect(addr, size, protect)
+        else:
+            for page in pages:
+                if page.committed:
+                    self.child.protect(page.addr, page.size, protect)
+
+        for page in pages:
+            page.protect = protect
+
+    def read(self, addr: int, size: int) -> bytearray:
+        #print(f"read({hex(addr)}, {hex(size)})")
+
+        pages = []
+        for page_addr, index, length in self.iter_chunks(addr, size):
+            page = self.pages.get(page_addr, None)
+            assert page is not None
+            pages.append((page, index, length))
+
+        if all([page.committed for page, _, _ in pages]):
+            return self.child.read(addr, size)
+        else:
+            data = bytearray(size)
+            for page, index, length in pages:
+                data_index = (page.addr + index) - addr
+                if page.committed:
+                    data[data_index:data_index + length] = self.child.read(page.addr + index, length)
+                else:
+                    if page.data is None:
+                        page.data = bytearray(page.size)
+                    data_chunk = page.data[index:index + length]
+                    data[data_index:data_index + length] = data_chunk
+            assert len(data) == size
+            return data
+
+    def write(self, addr: int, data: bytes) -> None:
+        #print(f"write({hex(addr)}, size: {hex(len(data))})")
+
+        pages = []
+        for page_addr, index, length in self.iter_chunks(addr, len(data)):
+            page = self.pages.get(page_addr, None)
+            assert page is not None
+            pages.append((page, index, length))
+
+        if all([page.committed for page, _, _ in pages]):
+            self.child.write(addr, data)
+        else:
+            for page, index, length in pages:
+                data_index = (page.addr + index) - addr
+                data_chunk = data[data_index:data_index + length]
+                assert len(data_chunk) == length
+                if page.committed:
+                    self.child.write(page.addr + index, data_chunk)
+                else:
+                    if page.data is None:
+                        page.data = bytearray(page.size)
+                    page.data[index:index + length] = data_chunk
+                    assert len(page.data) == page.size
+
+class SimpleTimer:
+    def __init__(self):
+        self.time = 0
+        self.start()
+
+    def start(self):
+        import time
+        self.time = time.perf_counter()
+
+    def __call__(self, name: str):
+        prev = self.time
+        self.start()
+        diff = self.time - prev
+        print(f"{name}: {diff*1000:.0f}ms")
+
 class Dumpulator(Architecture):
-    def __init__(self, minidump_file, *, trace=False, quiet=False, thread_id=None):
+    def __init__(self, minidump_file, *, trace=False, quiet=False, thread_id=None, debug_logs=False):
         self._quiet = quiet
-        self._debug = False
+        self._debug = debug_logs
 
         # Load the minidump
         self._minidump = minidump.MinidumpFile.parse(minidump_file)
@@ -103,19 +278,21 @@ class Dumpulator(Architecture):
         self.cs.detail = True
 
         self.regs = Registers(self._uc, self._x64)
-        self.args = Arguments(self._uc, self.regs, self._x64)
-        self.memory = MemoryManager(UnicornPageManager(self._uc))
+        self._pages = LazyPageManager(UnicornPageManager(self._uc))
+        self.memory = MemoryManager(self._pages)
+        self.args = Arguments(self._uc, self._pages, self.regs, self._x64)
         self.modules = ModuleManager(self.memory)
         self._allocate_base = None
         self._allocate_size = 1024 * 1024 * 10  # NOTE: 10 megs
         self._allocate_ptr = None
         self._setup_memory()
+        self.debug(f"total commit: {hex(self._pages.total_commit)}, pages: {self._pages.total_commit // PAGE_SIZE}")
         self._setup_modules()
+        self.syscalls = []
+        self._setup_syscalls()
         self._setup_emulator(thread)
         self.kill_me = None
         self.exit_code = None
-        self.syscalls = []
-        self._setup_syscalls()
         self.exports = self._all_exports()
         self.handles = HandleManager()
         self.exception = ExceptionInfo()
@@ -140,11 +317,12 @@ class Dumpulator(Architecture):
                 if region.state == MemoryState.MEM_RESERVE:
                     protect = region.allocation_protect
                 entry[3] = str(protect)
-                if isinstance(region.info, Module):
-                    module: Module = region.info
-                    entry[4] = f" {module.name}[{hex(module.size)}]"
-                elif region.info:
-                    entry[4] = str(region.info)
+                def pretty_info(info: Any):
+                    if isinstance(info, Module):
+                        return f"{info.name}[{hex(info.size)}]"
+                    else:
+                        return str(info)
+                entry[4] = ", ".join(map(pretty_info, region.info))
             table.append(entry)
         print(format_table(table))
 
@@ -181,9 +359,8 @@ class Dumpulator(Architecture):
 
     def _setup_gdt(self):
         # TODO: is the TSS actually necessary?
-        self._uc.mem_map(TSS_BASE, PAGE_SIZE, UC_PROT_READ | UC_PROT_WRITE)
-
-        self._uc.mem_map(GDT_BASE, PAGE_SIZE, UC_PROT_READ | UC_PROT_WRITE)
+        self._pages.commit(TSS_BASE, PAGE_SIZE, MemoryProtect.PAGE_READWRITE)
+        self._pages.commit(GDT_BASE, PAGE_SIZE, MemoryProtect.PAGE_READWRITE)
         for i in range(0, len(windows_gdt)):
             self.write(GDT_BASE + 8 * i, struct.pack("<Q", windows_gdt[i]))
         self.regs.gdtr = (0, GDT_BASE, 8 * len(windows_gdt) - 1, 0x0)
@@ -199,6 +376,8 @@ class Dumpulator(Architecture):
                 regions.append([])
             regions[-1].append(info)
         # NOTE: The HYPERVISOR_SHARED_DATA does not respect the allocation granularity
+        potential_hv = []
+        old_granularity = self.memory._granularity
         self.memory._granularity = PAGE_SIZE
         for i in range(len(regions)):
             region = regions[i]
@@ -218,14 +397,15 @@ class Dumpulator(Architecture):
             reserve_type = MemoryType(info.Type.value)
             self.debug(f" reserved: {hex(reserve_addr)}, size: {hex(reserve_size)}, protect: {reserve_protect}, type: {reserve_type}")
             self.memory.reserve(reserve_addr, reserve_size, reserve_protect, reserve_type)
+            if reserve_addr & (old_granularity - 1) != 0:
+                potential_hv.append(reserve_addr)
             for info in region:
                 emu_addr = info.BaseAddress & self.addr_mask
                 if info.State == minidump.MemoryState.MEM_COMMIT:
                     protect = reserve_protect if info.Protect is None else MemoryProtect(info.Protect.value)
                     self.debug(f"committed: {hex(emu_addr)}, size: {hex(info.RegionSize)}, protect: {protect}")
                     self.memory.commit(info.BaseAddress, info.RegionSize, protect)
-        self.memory._granularity = 0x10000
-
+        self.memory._granularity = old_granularity
         memory = self._minidump.get_reader().get_buffered_reader()
         seg: minidump.MinidumpMemorySegment
         for seg in self._minidump.memory_segments_64.memory_segments:
@@ -234,22 +414,124 @@ class Dumpulator(Architecture):
             memory.move(seg.start_virtual_address)
             assert memory.current_position == seg.start_virtual_address
             data = memory.read(seg.size)
-            self._uc.mem_write(emu_addr, data)
+            self._pages.write(emu_addr, data)
+        self._pages.lazy = False
+
+        self.memory.set_region_info(0x7ffe0000, "KUSER_SHARED_DATA")
+        if len(potential_hv) == 1:
+            self.memory.set_region_info(potential_hv[0], "HYPERVISOR_SHARED_DATA")
+        elif len(potential_hv) > 1:
+            self.debug(f"Unexpected unaligned addresses: {' '.join([hex(x) for x in potential_hv])}")
+
+    def _setup_pebteb(self, thread):
+        self.teb = thread.Teb & 0xFFFFFFFFFFFFF000
+        is_wow64 = self.modules["ntdll.dll"].find_export("Wow64Transition") is not None
+
+        for i in range(0, len(self._minidump.threads.threads)):
+            thread = self._minidump.threads.threads[i]
+            teb = thread.Teb & 0xFFFFFFFFFFFFF000
+            tid = thread.ThreadId
+            if self._x64:
+                # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_NT_TIB
+                stack_base = self.read_ptr(teb + 0x8)
+                stack_limit = self.read_ptr(teb + 0x10)
+                deallocation_stack = self.read_ptr(teb + 0x1478)
+            else:
+                # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_NT_TIB
+                stack_base = self.read_ptr(teb + 0x4)
+                stack_limit = self.read_ptr(teb + 0x8)
+                deallocation_stack = self.read_ptr(teb + 0xe0c)
+            # The stack grows from base (the higher address) to limit (the lower address)
+            self.memory.set_region_info(stack_base - 1, f"Stack (thread {tid})")
+
+            teb_size = 2 * PAGE_SIZE
+            self.memory.set_region_info(teb, f"TEB (thread {tid})", size=teb_size)
+            if is_wow64:
+                self.memory.set_region_info(teb - teb_size, f"WoW64 TEB (thread {tid})", size=teb_size)
+
+        # https://en.wikipedia.org/wiki/Win32_Thread_Information_Block
+        # Handle PEB
+        # Retrieve console handle
+        if self._x64:
+            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_TEB
+            self.peb = self.read_ptr(self.teb + 0x60)
+            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_PEB
+            process_parameters = self.read_ptr(self.peb + 0x20)
+            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_RTL_USER_PROCESS_PARAMETERS
+            self.console_handle = self.read_ptr(process_parameters + 0x10)
+            self.stdin_handle = self.read_ptr(process_parameters + 0x20)
+            self.stdout_handle = self.read_ptr(process_parameters + 0x28)
+            self.stderr_handle = self.read_ptr(process_parameters + 0x30)
+            number_of_heaps = self.read_ulong(self.peb + 0xe8)
+            process_heaps_ptr = self.read_ptr(self.peb + 0xf0)
+            api_set_map = self.read_ptr(self.peb + 0x68)
+            csr_shared_memory = self.read_ptr(self.peb + 0x88)
+            codepage_data = self.read_ptr(self.peb + 0xa0)
+            gdi_handle_table = self.read_ptr(self.peb + 0xf8)
+            shim_data = self.read_ptr(self.peb + 0x2d8)
+            activation_context_data = self.read_ptr(self.peb + 0x2f8)
+            default_activation_context_data = self.read_ptr(self.peb + 0x308)
+            leap_second_data = self.read_ptr(self.peb + 0x7b8)
+        else:
+            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_TEB
+            self.peb = self.read_ptr(self.teb + 0x30)
+            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_PEB
+            process_parameters = self.read_ptr(self.peb + 0x10)
+            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_RTL_USER_PROCESS_PARAMETERS
+            self.console_handle = self.read_ptr(process_parameters + 0x10)
+            self.stdin_handle = self.read_ptr(process_parameters + 0x18)
+            self.stdout_handle = self.read_ptr(process_parameters + 0x1c)
+            self.stderr_handle = self.read_ptr(process_parameters + 0x20)
+            number_of_heaps = self.read_ulong(self.peb + 0x88)
+            process_heaps_ptr = self.read_ptr(self.peb + 0x90)
+            api_set_map = self.read_ptr(self.peb + 0x38)
+            csr_shared_memory = self.read_ptr(self.peb + 0x4c)
+            codepage_data = self.read_ptr(self.peb + 0x58)
+            gdi_handle_table = self.read_ptr(self.peb + 0x94)
+            shim_data = self.read_ptr(self.peb + 0x1e8)
+            activation_context_data = self.read_ptr(self.peb + 0x1f8)
+            default_activation_context_data = self.read_ptr(self.peb + 0x200)
+            leap_second_data = self.read_ptr(self.peb + 0x470)
+
+        self.memory.set_region_info(self.peb, "PEB", size=PAGE_SIZE)
+        if is_wow64:
+            self.memory.set_region_info(self.peb - PAGE_SIZE, "WoW64 PEB", size=PAGE_SIZE)
+
+        self.info(f"TEB: 0x{self.teb:x}, PEB: 0x{self.peb:x}")
+        self.info(f"  ConsoleHandle: 0x{self.console_handle:x}")
+        self.info(f"  StandardInput: 0x{self.stdin_handle:x}")
+        self.info(f"  StandardOutput: 0x{self.stdout_handle:x}")
+        self.info(f"  StandardError: 0x{self.stderr_handle:x}")
+
+        process_heaps = []
+        for i in range(0, min(number_of_heaps, 0x1000)):
+            heap_ptr = self.read_ptr(process_heaps_ptr + self.ptr_size() * i)
+            process_heaps.append(heap_ptr)
+            self.memory.set_region_info(heap_ptr, f"Heap (ID {i})")
+
+        self.memory.set_region_info(api_set_map, "ApiSetMap")
+        self.memory.set_region_info(csr_shared_memory, "CSR shared memory")
+        self.memory.set_region_info(codepage_data, "CodePage data")
+        self.memory.set_region_info(gdi_handle_table, "GDI shared handle table")
+        self.memory.set_region_info(shim_data, "Shim data")
+        self.memory.set_region_info(activation_context_data, "Activation context data")
+        self.memory.set_region_info(default_activation_context_data, "Default activation context data")
+        self.memory.set_region_info(leap_second_data, "Leap second data")
 
     def _setup_emulator(self, thread):
+        self._setup_pebteb(thread)
         # TODO: map these using self.memory instead
         # map in codecaves (TODO: can be mapped as UC_PROT_NONE unless used)
-        self._uc.mem_map(USER_CAVE, PAGE_SIZE)
-        self._uc.mem_write(USER_CAVE, b"\xCC" * PAGE_SIZE)
-        self._uc.mem_map(KERNEL_CAVE, PAGE_SIZE)
+        self._pages.commit(USER_CAVE, PAGE_SIZE, MemoryProtect.PAGE_EXECUTE_WRITECOPY)
+        self._pages.write(USER_CAVE, b"\xCC" * PAGE_SIZE)
+        self._pages.commit(KERNEL_CAVE, PAGE_SIZE, MemoryProtect.PAGE_EXECUTE_WRITECOPY)
         kernel_code = bytearray(b"\xCC" * (PAGE_SIZE // 2) + b"\x00" * (PAGE_SIZE // 2))
         kernel_code[IRETQ_OFFSET] = 0x48
         kernel_code[IRETD_OFFSET] = 0xCF
-        self._uc.mem_write(KERNEL_CAVE, bytes(kernel_code))
+        self._pages.write(KERNEL_CAVE, bytes(kernel_code))
 
         # Set up context
         self._setup_gdt()
-        self.teb = thread.Teb & 0xFFFFFFFFFFFFF000
         if self._x64:
             self.regs.cs = windows_user_segment.cs
             self.regs.ss = windows_user_segment.ss
@@ -344,35 +626,6 @@ class Dumpulator(Architecture):
         if self.trace:
             self._uc.hook_add(UC_HOOK_CODE, _hook_code, user_data=self)
 
-        # https://en.wikipedia.org/wiki/Win32_Thread_Information_Block
-        # Handle PEB
-        # Retrieve console handle
-        if self._x64:
-            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_TEB
-            self.peb = self.read_ptr(self.teb + 0x60)
-            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_PEB
-            process_parameters = self.read_ptr(self.peb + 0x20)
-            # https://www.vergiliusproject.com/kernels/x64/Windows%2011/21H2%20(RTM)/_RTL_USER_PROCESS_PARAMETERS
-            self.console_handle = self.read_ptr(process_parameters + 0x10)
-            self.stdin_handle = self.read_ptr(process_parameters + 0x20)
-            self.stdout_handle = self.read_ptr(process_parameters + 0x28)
-            self.stderr_handle = self.read_ptr(process_parameters + 0x30)
-        else:
-            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_TEB
-            self.peb = self.read_ptr(self.teb + 0x30)
-            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_PEB
-            process_parameters = self.read_ptr(self.peb + 0x10)
-            # https://www.vergiliusproject.com/kernels/x86/Windows%2010/2110%2021H2%20(November%202021%20Update)/_RTL_USER_PROCESS_PARAMETERS
-            self.console_handle = self.read_ptr(process_parameters + 0x10)
-            self.stdin_handle = self.read_ptr(process_parameters + 0x18)
-            self.stdout_handle = self.read_ptr(process_parameters + 0x1c)
-            self.stderr_handle = self.read_ptr(process_parameters + 0x20)
-        self.info(f"TEB: 0x{self.teb:x}, PEB: 0x{self.peb:x}")
-        self.info(f"  ConsoleHandle: 0x{self.console_handle:x}")
-        self.info(f"  StandardInput: 0x{self.stdin_handle:x}")
-        self.info(f"  StandardOutput: 0x{self.stdout_handle:x}")
-        self.info(f"  StandardError: 0x{self.stderr_handle:x}")
-
     def _all_exports(self):
         exports: Dict[int, str] = {}
         for module in self.modules:
@@ -472,12 +725,18 @@ class Dumpulator(Architecture):
         self.regs.csp = csp
 
     def read(self, addr, size):
-        return self._uc.mem_read(addr, size)
+        if not isinstance(addr, int):
+            addr = int(addr)
+        return self._pages.read(addr, size)
 
     def write(self, addr, data):
-        self._uc.mem_write(addr, data)
+        if not isinstance(addr, int):
+            addr = int(addr)
+        self._pages.write(addr, data)
 
     def call(self, addr, args: List[int] = [], regs: dict = {}, count=0):
+        if not isinstance(addr, int):
+            addr = int(addr)
         # allow passing custom registers
         for name, value in regs.items():
             self.regs.__setattr__(name, value)
@@ -498,6 +757,7 @@ class Dumpulator(Architecture):
     def allocate(self, size, page_align=False):
         if not self._allocate_ptr:
             self._allocate_base = self.memory.find_free(self._allocate_size)
+            assert self._allocate_base is not None, "Failed to find free memory"
             self.memory.reserve(
                 start=self._allocate_base,
                 size=self._allocate_size,
@@ -825,7 +1085,7 @@ rsp in KiUserExceptionDispatcher:
 
     def load_dll(self, file_name: str, file_data: bytes):
         self.handles.map_file("\\??\\" + file_name, FileObject(file_name, file_data))
-        argument_ptr = self.allocate(0x1000)
+        argument_ptr = self.allocate(PAGE_SIZE)
         utf16 = file_name.encode("utf-16-le")
         if self._x64:
             argument_data = struct.pack("<IIQHHIQ", 0, 0, 0, len(utf16), len(utf16) + 2, 0, argument_ptr + 32)
@@ -858,6 +1118,10 @@ def _hook_code_exception(uc: Uc, address, size, dp: Dumpulator):
         raise err
 
 def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
+    if dp._pages.handle_lazy_page(address, min(size, PAGE_SIZE)):
+        dp.debug(f"committed lazy page {hex(address)}[{hex(size)}]")
+        return True
+
     fetch_accesses = [UC_MEM_FETCH, UC_MEM_FETCH_PROT, UC_MEM_FETCH_UNMAPPED]
     if access == UC_MEM_FETCH_UNMAPPED and address >= FORCE_KILL_ADDR - 0x10 and address <= FORCE_KILL_ADDR + 0x10 and dp.kill_me is not None:
         dp.error(f"forced exit memory operation {access} of {address:x}[{size:x}] = {value:X}")
