@@ -1,5 +1,5 @@
 from enum import Enum, IntFlag, Flag
-from typing import Any, List, Dict, Union
+from typing import Any, List, Dict, Union, Optional, Set
 import bisect
 
 PAGE_SIZE = 0x1000
@@ -86,8 +86,7 @@ class MemoryRegion:
         return f"MemoryRegion({hex(self.start)}, {hex(self.size)}, {self.protect}, {self.type}, {repr(self.info)})"
 
     def pages(self):
-        for page in range(self.start, self.end, PAGE_SIZE):
-            yield page
+        return range(self.start, self.end, PAGE_SIZE)
 
 class PageManager:
     def commit(self, addr: int, size: int, protect: MemoryProtect) -> None:
@@ -99,6 +98,12 @@ class PageManager:
     def protect(self, addr: int, size: int, protect: MemoryProtect) -> None:
         raise NotImplementedError()
 
+    def read(self, addr: int, size: int) -> bytearray:
+        raise NotImplementedError()
+
+    def write(self, addr: int, data: bytes) -> None:
+        raise NotImplementedError()
+
 class MemoryBasicInformation:
     def __init__(self, base: int, allocation_base: int, allocation_protect: MemoryProtect):
         self.base = base
@@ -108,13 +113,13 @@ class MemoryBasicInformation:
         self.state: MemoryState = None
         self.protect: MemoryProtect = None
         self.type: MemoryType = None
-        self.info: Any = None
+        self.info: List[Any] = []
 
     def __str__(self):
         return f"MemoryBasicInformation(base: {hex(self.base)}, allocation_base: {hex(self.allocation_base)}, region_size: {hex(self.region_size)}, state: {self.state}, protect: {self.protect}, type: {self.type})"
 
 class MemoryManager:
-    def __init__(self, page_manager: PageManager, minimum = 0x10000, maximum = 0x7fffffff0000, granularity = 0x10000):
+    def __init__(self, page_manager: PageManager, minimum=0x10000, maximum=0x7fffffff0000, granularity=0x10000):
         self._page_manager = page_manager
         self._minimum = minimum
         self._maximum = maximum
@@ -122,9 +127,9 @@ class MemoryManager:
         self._regions: List[MemoryRegion] = []
         self._committed: Dict[int, MemoryRegion] = {}
 
-    def find_region(self, region: Union[MemoryRegion, int]):
+    def find_region(self, region: Union[MemoryRegion, int]) -> Optional[MemoryRegion]:
         if isinstance(region, int):
-            region = MemoryRegion(self.align_page(region), 0)
+            region = MemoryRegion(self.containing_page(region), 0)
         index = bisect.bisect_right(self._regions, region)
         if index == 0:
             return None
@@ -135,30 +140,43 @@ class MemoryManager:
             else:
                 return None
 
-    def find_commit(self, addr: int):
-        addr = self.align_page(addr)
+    def find_commit(self, addr: int) -> Optional[MemoryRegion]:
+        addr = self.containing_page(addr)
         return self._committed.get(addr, None)
 
-    def align_page(self, addr: int):
+    # Rounds down to the page containing this address
+    def containing_page(self, addr: int) -> int:
+        mask = PAGE_SIZE - 1
+        return addr & ~mask
+
+    # Rounds up to the nearest page size
+    def align_page(self, addr: int) -> int:
         mask = PAGE_SIZE - 1
         return (addr + mask) & ~mask
 
-    def align_allocation(self, addr: int):
+    # Rounds up to the nearest allocation granularity
+    def align_allocation(self, addr: int) -> int:
         mask = self._granularity - 1
         return (addr + mask) & ~mask
 
-    def find_free(self, size: int):
+    def find_free(self, size: int, allocation_align=True) -> Optional[int]:
         assert size > 0 and self.align_page(size) == size
         base = self._minimum
         while base < self._maximum:
             info = self.query(base)
             assert info.base == base
-            if info.state == MemoryState.MEM_FREE and info.region_size >= size and self.align_allocation(base) == base:
-                return info.base
             base += info.region_size
+            if info.state == MemoryState.MEM_FREE:
+                if allocation_align:
+                    aligned_base = self.align_allocation(info.base)
+                    diff = aligned_base - info.base
+                    info.base = aligned_base
+                    info.region_size -= diff
+                if info.region_size >= size:
+                    return info.base
         return None
 
-    def reserve(self, start: int, size: int, protect: MemoryProtect, type: MemoryType = MemoryType.MEM_PRIVATE, info: Any = None):
+    def reserve(self, start: int, size: int, protect: MemoryProtect, type: MemoryType = MemoryType.MEM_PRIVATE, info: Any = None) -> None:
         assert isinstance(protect, MemoryProtect)
         assert isinstance(type, MemoryType)
         assert size > 0 and self.align_page(size) == size
@@ -181,7 +199,7 @@ class MemoryManager:
             check_overlaps(index)
         self._regions.insert(index, region)
 
-    def release(self, start: int):
+    def release(self, start: int) -> None:
         assert self.align_allocation(start) == start
 
         parent_region = self.find_region(start)
@@ -204,10 +222,10 @@ class MemoryManager:
         assert parent_region.commit_count == 0
         self._regions.remove(parent_region)
 
-    def commit(self, start: int, size: int, protect: MemoryProtect = MemoryProtect.UNDEFINED):
+    def commit(self, start: int, size: int, protect: MemoryProtect = MemoryProtect.UNDEFINED) -> None:
         assert isinstance(protect, MemoryProtect)
         assert size > 0 and self.align_page(size) == size
-        assert self.align_page(start) == start
+        assert self.containing_page(start) == start
         region = MemoryRegion(start, size)
         parent_region = self.find_region(region)
         if parent_region is None:
@@ -234,9 +252,9 @@ class MemoryManager:
                     self._committed[page] = MemoryRegion(page, PAGE_SIZE, protect, parent_region.type)
                     parent_region.commit_count += 1
 
-    def decommit(self, start: int, size: int):
+    def decommit(self, start: int, size: int) -> None:
         assert size > 0 and self.align_page(size) == size
-        assert self.align_page(start) == start
+        assert self.containing_page(start) == start
         region = MemoryRegion(start, size)
         parent_region = self.find_region(region)
         if parent_region is None:
@@ -254,10 +272,10 @@ class MemoryManager:
                     del self._committed[page]
                     parent_region.commit_count -= 1
 
-    def protect(self, start: int, size: int, protect: MemoryProtect):
+    def protect(self, start: int, size: int, protect: MemoryProtect) -> MemoryProtect:
         assert isinstance(protect, MemoryProtect)
         assert size > 0 and self.align_page(size) == size
-        assert self.align_page(start) == start
+        assert self.containing_page(start) == start
         region = MemoryRegion(start, size)
         parent_region = self.find_region(region)
         if parent_region is None:
@@ -278,8 +296,8 @@ class MemoryManager:
 
         return old_protect
 
-    def query(self, start: int):
-        start = self.align_page(start)
+    def query(self, start: int) -> MemoryBasicInformation:
+        start = self.containing_page(start)
 
         region = MemoryRegion(start, 0)
         parent_region = self.find_region(region)
@@ -298,20 +316,25 @@ class MemoryManager:
 
         # Reference: https://learn.microsoft.com/en-us/windows/win32/api/memoryapi/nf-memoryapi-virtualquery#remarks
         result: MemoryBasicInformation = None
+        result_info = {}
+        def add_info(region: MemoryRegion):
+            if region.info is None:
+                return
+            if region.info in result_info:
+                return
+            result_info[region.info] = region.start
         for page in parent_region.pages():
             if page < start:
                 continue
             elif result is None:
                 result = MemoryBasicInformation(page, parent_region.start, parent_region.protect)
-                if page == parent_region.start:
-                    result.info = parent_region.info
+                add_info(parent_region)
                 if page in self._committed:
                     result.state = MemoryState.MEM_COMMIT
                     commited_page = self._committed[page]
                     result.protect = commited_page.protect
                     result.type = commited_page.type
-                    if commited_page.info:
-                        result.info = commited_page.info
+                    add_info(commited_page)
                     assert commited_page.type == parent_region.type
                 else:
                     result.state = MemoryState.MEM_RESERVE
@@ -330,17 +353,53 @@ class MemoryManager:
                 elif result.state == MemoryState.MEM_COMMIT:
                     if commited_page is not None and commited_page.type == result.type and commited_page.protect == result.protect:
                         result.region_size += PAGE_SIZE
+                        add_info(commited_page)
                     else:
                         break
                 else:
                     assert False  # unreachable
+        # Only keep information starting from the current page, or the parent page if none
+        if result is not None:
+            result.info = []
+            for info, start_addr in result_info.items():
+                if start_addr >= result.base:
+                    result.info.append(info)
+            if len(result.info) == 0 and len(result_info) > 0:
+                result.info = list(result_info.keys())[:1]
+
         return result
 
-    def map(self):
+    def map(self) -> List[MemoryBasicInformation]:
         addr = self._minimum
         regions: List[MemoryBasicInformation] = []
         while addr < self._maximum:
             info = self.query(addr)
             regions.append(info)
             addr += info.region_size
-        return regions
\ No newline at end of file
+        return regions
+
+    def read(self, addr: int, size: int) -> bytearray:
+        return self._page_manager.read(addr, size)
+
+    def write(self, addr: int, data: bytes):
+        return self._page_manager.write(addr, data)
+
+    def set_region_info(self, addr: int, info: Any, *, size=0):
+        region = self.find_region(addr)
+        if region is None:
+            return False
+
+        if size > 0:
+            for page in region.pages():
+                if page < addr:
+                    continue
+                if page >= addr + size:
+                    break
+                commit = self._committed.get(page)
+                if commit is not None and commit.info is None:
+                    commit.info = info
+        else:
+            if region.info is not None:
+                return False
+            region.info = info
+        return True
